#!/usr/bin/env python3
"""
Module GROUP -- UI Button Group Widgets
Sub-Package UI.BASE of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

from plib.ui.defs import *
from plib.ui.coll import BaseStringListWidget
from plib.ui.common import font_args


class PButtonGroupMixin(BaseStringListWidget):
    # To be used as a mixin class for button groups in each toolkit,
    # subclassed from PPanel
    
    signals = (
        SIGNAL_BUTTONSELECTED,
    )
    
    def __init__(self, items=None, value=None, starting_index=None, font=None):
        self.item_font = font
        self.init_group()
        BaseStringListWidget.__init__(self, items, value, starting_index)
    
    def init_group(self):
        pass  # Toolkits may override to set up internal group object
    
    def create_button(self, item):
        """Derived classes must implement.
        """
        raise NotImplementedError
    
    def set_button_font(self, button, font_name=None, font_size=None, bold=None, italic=None):
        """Derived classes must implement.
        """
        raise NotImplementedError
    
    def add_button(self, item):
        widget = self.create_button(item)
        if self.item_font:
            self.set_button_font(widget, *self.item_font)
        self.add_widget(widget)
        self.contents.append(widget)
    
    def init_button(self, button):
        """Derived classes must implement.
        """
        raise NotImplementedError
    
    def add_widget(self, widget):
        self.init_button(widget)
        # This should call the toolkit's PPanel.add_widget
        super(PButtonGroupMixin, self).add_widget(widget)
    
    def group_buttons(self):
        """Derived classes must implement.
        """
        raise NotImplementedError
    
    @property
    def buttons(self):
        return self.group_buttons()
    
    def del_button(self, button):
        pass  # Subclasses may override
    
    def remove_button(self, button):
        self.contents.remove(button)
        self.remove_widget(button)
        self.del_button(button)
    
    def is_checked(self, button):
        """Derived classes must implement.
        """
        raise NotImplementedError
    
    def check_button(self, button, checked=True):
        """Derived classes must implement.
        """
        raise NotImplementedError
    
    def button_text(self, button):
        """Derived classes must implement.
        """
        raise NotImplementedError
    
    def set_button_text(self, button, value):
        """Derived classes must implement.
        """
        raise NotImplementedError
    
    def button_text_at(self, index):
        return self.button_text(self.buttons[index])
    
    def set_button_text_at(self, index, value):
        self.set_button_text(self.buttons[index], value)
    
    def selected_button(self):
        return first(button for button in self.buttons if self.is_checked(button))
    
    def current_text(self):
        button = self.selected_button()
        return self.button_text(button) if button else None
    
    def current_index(self):
        return first(i for i, button in enumerate(self.buttons) if self.is_checked(button))
    
    def set_current_index(self, index):
        self.check_button(self.buttons[index])
    
    def on_button_toggled(self, button, checked):
        if checked:
            index = self.buttons.index(button)
            self.do_notify(SIGNAL_BUTTONSELECTED, index)
    
    # Sequence emulator methods; note that the sequence interface does not support
    # inserting buttons at arbitrary indexes, only appending buttons and removing
    # them
    
    def _indexlen(self):
        return len(self.buttons)  # toolkits may override if a faster method exists
    
    def _get_data(self, index):
        return self.button_text_at(index)
    
    def _set_data(self, index, value):
        self.set_button_text_at(index, value)
    
    def _add_data(self, index, value):
        if index != len(self):
            raise ValueError("Cannot insert buttons in middle of button group")
        self.add_button(value)
    
    def _del_data(self, index):
        self.remove_button(self.contents[index])



#!/usr/bin/env python3
"""
Module DEFS -- Common UI Definitions
Sub-Package UI of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

Defines common constants and functions used by various UI modules.
"""

import sys

# general constants
DEFAULT_FONT_SIZE = 10

# constants for referring to UI toolkits
#UI_QT = 1
#UI_GTK = 2
UI_WX = 3
#UI_KDE = 4
#UI_QT4 = 5
#UI_KDE4 = 6
#UI_PYSIDE = 7
UI_QT5 = 8
UI_PYSIDE2 = 9
#UI_WEB = 10

# message box types
MBOX_INFO = 0
MBOX_WARN = 1
MBOX_ERROR = 2
MBOX_QUERY = 3

# constants for message box responses
ANSWER_NONE = 0
ANSWER_YES = 1
ANSWER_NO = 2
ANSWER_CANCEL = 3
ANSWER_OK = 4

# constants for alignment
ALIGN_LEFT = 1
ALIGN_CENTER = 2
ALIGN_RIGHT = 3
ALIGN_TOP = 4
ALIGN_BOTTOM = 5
ALIGN_JUST = 9

# Layout constants
LAYOUT_HORIZONTAL = 1
LAYOUT_VERTICAL = 2
LAYOUT_COLGRID = 3
LAYOUT_ROWGRID = 4

# Panel style constants
PANEL_PLAIN = 0
PANEL_BOX = 1
PANEL_RAISED = 2
PANEL_SUNKEN = 3

# Prefs dialog section style constants
SECTION_TAB = 0  # this is the default
SECTION_GROUPBOX = 1

# constants for signals
SIGNAL_ACTIVATED = 10  # widget has received focus
SIGNAL_CLICKED = 101  # widget has been clicked
SIGNAL_TOGGLED = 102  # on/off widget has been toggled
SIGNAL_SELECTED = 151  # item in widget has been selected; handler must take index param
SIGNAL_BUTTONSELECTED = 156  # item in button group has been selected; handler must take index param
SIGNAL_LISTVIEWSELECTED = 161  # list view item has changed; handler must take item param
SIGNAL_LISTBOXSELECTED = 165  # list box item has changed; handler must take index param
SIGNAL_CELLSELECTED = 171  # table current cell has changed; handler must take row, col params
SIGNAL_CELLCHANGED = 301  # table cell text has changed; handler must take row, col params for changed cell
SIGNAL_TEXTCHANGED = 401  # text in edit control has been changed
SIGNAL_TEXTMODCHANGED = 406  # modification state of edit control has changed
SIGNAL_TEXTSTATECHANGED = 411  # state of edit control has changed
SIGNAL_EDITCHANGED = 450  # text in edit box has been changed
SIGNAL_NUMEDITCHANGED = 460  # text in edit box has been changed
SIGNAL_VALUECHANGED = 470  # text in edit box has been changed
SIGNAL_ENTER = 490  # enter/return key has been pressed while widget has focus
SIGNAL_TABSELECTED = 501  # new tab has been selected; handler must take int param for new index
SIGNAL_FINISHED = 601  # dialog is finished
SIGNAL_QUERYCLOSE = 921  # widget is asking permission to close
SIGNAL_CLOSING = 931  # widget close has been accepted
SIGNAL_SHOWN = 941  # widget has been shown
SIGNAL_HIDDEN = 951  # widget has been hidden

# these signal constants are for internal use only
SIGNAL_MENUACTIVATED = 19  # internal menu/toolbar activation signal that gets re-interpreted
SIGNAL_KEYPRESSED = 499  # internal key pressed signal, filtered to SIGNAL_ENTER
SIGNAL_TABCURRENTCHANGED = 599  # internal tab changed signal that gets re-interpreted
SIGNAL_NOTIFIER = 801  # socket notifier has received an event notification
SIGNAL_TABLECELLSELECTED = 829  # internal table cell selected signal that gets re-interpreted
SIGNAL_TABLECHANGED = 831  # internal table changed event that gets re-interpreted
SIGNAL_TABLECELLEDITING = 835  # internal table cell edit event to track changes
SIGNAL_TABLECELLEDITDONE = 839  # internal table cell edit event to track changes
SIGNAL_WIDGETCHANGED = 901  # widget has been changed (not including above specific changes)
SIGNAL_SIZEEVENT = 935  # internal window size event that gets re-interpreted
SIGNAL_SHOWEVENT = 945  # internal window show/hide event that gets re-interpreted
SIGNAL_MINIMIZEEVENT = 955  # internal window minimize event that gets re-interpreted
SIGNAL_CLOSEEVENT = 989  # internal main window close event that gets re-interpreted
SIGNAL_BEFOREQUIT = 999  # app is about to quit

# these signal constants are for custom widget signal mixins
SIGNAL_FOCUS_IN = 191  # widget has received keyboard focus
SIGNAL_FOCUS_OUT = 195  # widget has lost keyboard focus
SIGNAL_LEFTCLICK = 201  # widget has received left mouse click
SIGNAL_RIGHTCLICK = 205  # widget has received right mouse click
SIGNAL_MIDDLECLICK = 207  # widget has received middle mouse click
SIGNAL_LEFTDBLCLICK = 211  # widget has received left mouse double click
SIGNAL_RIGHTDBLCLICK = 215  # widget has received right mouse double click
SIGNAL_MIDDLEDBLCLICK = 217  # widget has received middle mouse double click

# constants for action flags, used as keys
ACTION_FILE_NEW = 5
ACTION_FILE_OPEN = 10
ACTION_FILE_SAVE = 15
ACTION_FILE_SAVEAS = 20
ACTION_FILE_CLOSE = 25
ACTION_EDIT_UNDO = 260
ACTION_EDIT_REDO = 270
ACTION_EDIT_CUT = 310
ACTION_EDIT_COPY = 320
ACTION_EDIT_PASTE = 330
ACTION_EDIT_DELETE = 340
ACTION_EDIT_SELECTALL = 350
ACTION_EDIT_SELECTNONE = 355
ACTION_EDIT_CLEAR = 360
ACTION_EDIT_OVERWRITE = 390
ACTION_VIEW = 500
ACTION_EDIT = 505
ACTION_OK = 550
ACTION_CANCEL = 660
ACTION_REFRESH = 880
ACTION_ADD = 1024
ACTION_REMOVE = 2048
ACTION_APPLY = 8192
ACTION_COMMIT = 8288
ACTION_ROLLBACK = 8298
ACTION_PREFS = 9000
ACTION_ABOUT = 9152
ACTION_ABOUT_TOOLKIT = 9160
ACTION_EXIT = 9999

# Menu titles

MENU_FILE = "&File"
MENU_EDIT = "&Edit"
MENU_ACTION = "&Action"
MENU_OPTIONS = "&Options"
MENU_HELP = "&Help"

# color constants -- we choose values to make hacks easier :)
COLORNAMES = [
    'BLACK',
    'DARKRED', 'RED',  # 'LIGHTRED', # FIXME: for some reason this isn't in the standard X11 list
    'DARKGREEN', 'GREEN', 'LIGHTGREEN',
    'DARKBLUE', 'BLUE', 'LIGHTBLUE',
    'YELLOW', 'MAGENTA', 'CYAN',
    'DARKGRAY', 'GRAY', 'LIGHTGRAY', 'WHITE'
]
for color in COLORNAMES:
    setattr(sys.modules[__name__], 'COLOR_{}'.format(color), color)

# top window geometry constants
SIZE_NONE = 0
SIZE_CLIENTWRAP = 1
SIZE_MAXIMIZED = 2
SIZE_OFFSET = 4
SIZE_SETTINGS = 8
SIZE_DESKTOP = 16

MOVE_NONE = 0
MOVE_CENTER = 1
MOVE_SETTINGS = 2

# Socket notifier constants
NOTIFY_READ = 0
NOTIFY_WRITE = 1

# Control variables for widget geometries

FRAME_MARGIN = 10
TAB_MARGIN = 10
PANEL_MARGIN = 4
BOX_MARGIN = 1
PANEL_SPACING = 10
BUTTON_WIDTH = 110

# Width values for headers; a number > 0 means set to that width

WIDTH_CONTENTS = 0
WIDTH_STRETCH = -1

# Convenience function to add or mutate constants; must
# be called before any other module besides this one is
# imported

def add_constants(**kw):
    for name, value in kw.items():
        setattr(sys.modules[__name__], name, value)


# Maps for widget construction and event handler connection

WIDGET_EVENT_MAP = {
    SIGNAL_ACTIVATED: None,
    SIGNAL_CLICKED: 'clicked',
    SIGNAL_TOGGLED: 'toggled',
    SIGNAL_SELECTED: 'selected',
    SIGNAL_BUTTONSELECTED: 'selected',
    SIGNAL_LISTVIEWSELECTED: 'selected',
    SIGNAL_LISTBOXSELECTED: 'selected',
    SIGNAL_CELLSELECTED: 'selected',
    SIGNAL_CELLCHANGED: 'changed',
    SIGNAL_TEXTCHANGED: 'changed',
    SIGNAL_TEXTMODCHANGED: 'mod_changed',
    SIGNAL_TEXTSTATECHANGED: 'state_changed',
    SIGNAL_EDITCHANGED: 'changed',
    SIGNAL_NUMEDITCHANGED: 'value_changed',
    SIGNAL_VALUECHANGED: 'value_changed',
    SIGNAL_ENTER: 'enter',
    SIGNAL_TABSELECTED: 'selected',
    SIGNAL_QUERYCLOSE: 'queryclose',
    SIGNAL_CLOSING: 'closing',
    SIGNAL_SHOWN: 'shown',
    SIGNAL_HIDDEN: 'hidden',
    
    SIGNAL_FOCUS_IN: 'focus_in',
    SIGNAL_FOCUS_OUT: 'focus_out',
    SIGNAL_LEFTCLICK: 'left_click',
    SIGNAL_RIGHTCLICK: 'right_click',
    SIGNAL_MIDDLECLICK: 'middle_click',
    SIGNAL_LEFTDBLCLICK: 'left_dblclick',
    SIGNAL_RIGHTDBLCLICK: 'right_dblclick',
    SIGNAL_MIDDLEDBLCLICK: 'middle_dblclick',
}

add_handlers = WIDGET_EVENT_MAP.update

WIDGET_ATTR_PREFIXES = {
    # Only widgets whose prefix is different from their module name need this
    'PCheckBox': 'checkbox',
    'PEditBox': 'edit',
    'PNumEditBox': 'edit',
    'PSpinEditBox': 'edit',
    'PPanel': 'panel',
    'PButtonGroup': 'buttongroup',
    'PRadioGroup': 'radiogroup',
    'PTreeView': 'treeview',
    'PEditControl': 'memo',
    'PTextDisplay': 'text',
    'PStatusLabel': 'label',
}

add_widget_prefixes = WIDGET_ATTR_PREFIXES.update

WIDGET_AUTO_PREFIXES = [
    # These prefixes are used in auto_name and should not be further prefixed
    'groupbox_',
    'padding_',
    'panel_',
    'labelbox_',
]

add_auto_prefix = WIDGET_AUTO_PREFIXES.append
add_auto_prefixes = WIDGET_AUTO_PREFIXES.extend



#!/usr/bin/env python3
"""
Module DIALOG -- UI Button Widget
Sub-Package UI.BASE of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

from plib.ui.defs import *
from plib.ui.widgets import widget_from_spec

from .app import PWidgetBase


class PDialogBase(PWidgetBase):
    """Base class for dialogs.
    """
    
    control_widget = None
    
    def __init__(self, manager, parent, caption, client,
                 accept_buttons=('ok', 'yes'), reject_buttons=('cancel', 'no')):
        
        PWidgetBase.__init__(self, manager, parent)
        self.set_caption(caption)
        self.create_controls(client)
        for button in accept_buttons:
            self.setup_button(button, self.accept_changes)
        for button in reject_buttons:
            self.setup_button(button, self.reject_changes)
    
    def setup_button(self, button, method):
        control = getattr(self.manager, 'button_{}'.format(button), None)
        if control:
            for sig in control.signals:
                control.setup_notify(sig, method)
    
    def get_controls_size(self):
        """Return tuple of (width, height) needed to wrap client.
        """
        c = self.control_widget
        return (c.preferred_width(), c.preferred_height())
    
    def create_controls(self, client):
        # The dialog itself has no layout capabilities, so there can
        # only be a single client: normally a panel or group box, so
        # the actual controls and layout are handled by the client
        self.control_widget = widget_from_spec(self.manager, self, client)
    
    def size_to_controls(self):
        self.set_size(*self.get_controls_size())
    
    def set_caption(self, caption):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def do_display(self):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def display(self):
        self.size_to_controls()
        self.do_display()
    
    def accept_changes(self):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def reject_changes(self):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError



#!/usr/bin/env python3
"""
Module CUSTOM -- Support for Custom Signals and Widgets
Sub-Package UI of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

Defines the API for adding custom signals and widgets.
"""

import sys


def set_toolkit(toolkit, err_msg=None, err_exit=True):
    from plib.ui import toolkits
    try:
        toolkits.set_toolkit(toolkit)
    except toolkits.ToolkitError:
        if err_msg:
            print(err_msg)
        if err_exit:
            sys.exit(1)
        raise


def add_signals(**kwargs):
    from plib.ui import defs
    from plib.ui import imp
    
    constants = {}
    handlers = {}
    custom_specs = []
    
    for name, specs in kwargs.items():
        signame = name.upper()
        assert signame.startswith("SIGNAL_")
        value, handler_name, *args = specs
        if isinstance(value, tuple):
            vname, offset = value
            value = getattr(defs, vname) + offset
        constants[signame] = value
        handlers[value] = handler_name
        custom_specs.append((signame,) + tuple(args))
    
    defs.add_constants(**constants)
    defs.add_handlers(handlers)
    
    # Have to do this after defs is updated above so custom signals get imported
    add_custom_signal = imp.get_toolkit_object('app', 'add_custom_signal')
    
    # Args are the types of arguments to be sent to the event handler
    # (Note: some toolkits ignore these)
    for args in custom_specs:
        add_custom_signal(*args)


def add_widget_prefixes(**kwargs):
    from plib.ui import defs
    
    defs.add_widget_prefixes(**kwargs)


def add_auto_prefixes(*args):
    from plib.ui import defs
    
    defs.add_auto_prefixes(args)


def add_action(key, name, caption, description, icon):
    from plib.ui import common
    if key in common.action_map:
        raise RuntimeError("Key {} is already in use".format(key))
    common.action_map[key] = [name, caption, description, icon]


def modify_action(key, name=None, caption=None, description=None, icon=None):
    from plib.ui import common
    entry = common.action_map.get(key)
    if not entry:
        raise RuntimeError("Invalid action key: {}".format(key))
    items = (name, caption, description, icon)
    add_index = len(entry)
    for index, item in enumerate(items[:add_index]):
        if item:
            entry[index] = item
    add_items = items[add_index:]
    if add_items:
        entry[add_index:] = add_items



#!/usr/bin/env python3
"""
Module COLL -- UI Collection Support
Sub-Package UI of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

from plib.stdlib.coll import baselist


class BaseListWidget(baselist):
    """Base class for widgets that support the list protocol.
    """
    
    def current_index(self):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def set_current_index(self, index):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError


class BaseStringListWidget(BaseListWidget):
    """Base class for widgets that look like a list of strings.
    """
    
    def __init__(self, items=None, value=None, starting_index=None):
        BaseListWidget.__init__(self, items)
        self.complete_init(items, value, starting_index)
    
    def complete_init(self, items, value, starting_index):
        if starting_index:
            self.set_current_index(starting_index)
        elif value:
            self.set_current_text(value)
        elif items:
            self.set_current_index(0)
    
    def add_items(self, items, value=None, starting_index=None):
        self.extend(items)
        self.complete_init(items, value, starting_index)
    
    def load_items(self, items, value=None, starting_index=None):
        self.clear()
        self.add_items(items, value, starting_index)
    
    def current_text(self):
        return self[self.current_index()]
    
    get_current_text = current_text
    
    def set_current_text(self, text):
        self.set_current_index(self.index(text))



#!/usr/bin/env python3
"""
Module BUTTON -- UI Button Widgets
Sub-Package UI.BASE of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

from plib.ui.defs import *

from .app import PActionMixin, PWidgetBase


class PButtonBase(PActionMixin, PWidgetBase):
    """Base class for push button.
    """
    
    signals = (
        SIGNAL_CLICKED,
    )
    
    fn_get_caption = None
    fn_set_caption = None
    
    def __init__(self, manager, parent, caption, icon=None,
                 geometry=None, font=None):
        
        if caption:
            self.caption = caption
        if icon:
            self.set_icon(icon)
        PWidgetBase.__init__(self, manager, parent,
                             geometry=geometry, font=font)
    
    def get_caption(self):
        return getattr(self, self.fn_get_caption)()
    
    def set_caption(self, value):
        getattr(self, self.fn_set_caption)(value)
    
    caption = property(get_caption, set_caption)
    
    def set_icon_obj(self, icon):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def set_icon(self, icon):
        if isinstance(icon, int):
            # Assume it's an action key
            icon = self.get_icon(icon)
        elif isinstance(icon, str):
            # Assume its an image filename
            icon = self.load_icon_from_file(icon)
        # At this point it should be an icon object
        self.set_icon_obj(icon)


class PActionButtonBase(PButtonBase):
    
    signals = (
        SIGNAL_ACTIVATED,
    )
    
    def __init__(self, manager, parent, action,
                 geometry=None, font=None):
        
        # These checks allow toolkits to override the captions and icons
        # by passing a zero or negative action key
        caption = self.get_menu_str(action) if (action or -1) > 0 else None
        icon = self.get_icon(action) if (action or -1) > 0 else None
        PButtonBase.__init__(self, manager, parent, caption, icon,
                             geometry=geometry, font=font)
        
        # Convert click events to activated events
        self.setup_notify(SIGNAL_CLICKED, self.activate)
    
    def activate(self):
        self.do_notify(SIGNAL_ACTIVATED)


class PCheckBoxBase(PButtonBase):
    """ Base class for checkbox.
    """
    
    signals = (
        SIGNAL_TOGGLED,
    )
    
    fn_get_checked = None
    fn_set_checked = None
    
    def __init__(self, manager, parent, label, checked=None,
                 geometry=None, font=None, tristate=False):
        
        PButtonBase.__init__(self, manager, parent, label,
                             geometry=geometry, font=font)
        if tristate:
            self.make_tristate()
        if checked is not None:
            self.checked = checked
    
    def make_tristate(self):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def get_checked(self):
        return getattr(self, self.fn_get_checked)()
    
    def set_checked(self, value):
        getattr(self, self.fn_set_checked)(value)
    
    checked = property(get_checked, set_checked)
    
    def set_icon(self, icon):
        # No icons on checkboxes
        pass



#!/usr/bin/env python3
"""
Module WX.FORM -- Python wxWidgets Panel Objects
Sub-Package UI.TOOLKITS.WX of Package PLIB3 -- Python GUI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the wxWidgets UI objects for the panel widgets.
"""

import wx

from plib.ui.defs import *
from plib.ui.base.form import PPanelBase

from .app import PWxWidget


wx_styles = {
    PANEL_PLAIN: wx.NO_BORDER,
    PANEL_BOX: wx.SIMPLE_BORDER,
    PANEL_RAISED: wx.RAISED_BORDER,
    PANEL_SUNKEN: wx.SUNKEN_BORDER,
}

wx_layouts = {
    LAYOUT_HORIZONTAL: (wx.BoxSizer, (wx.HORIZONTAL,)),
    LAYOUT_VERTICAL: (wx.BoxSizer, (wx.VERTICAL,)),
    LAYOUT_COLGRID: (wx.GridSizer, ()),
    LAYOUT_ROWGRID: (wx.GridSizer, ()),
}


class PPanel(PWxWidget, wx.Panel, PPanelBase):
    
    _expand = True  # used by parent panel to determine placement
    
    _is_grid = False
    
    def __init__(self, manager, parent, align=ALIGN_JUST, layout=LAYOUT_VERTICAL, contents=None,
                 style=PANEL_PLAIN, margin=None, spacing=None):
        
        wx.Panel.__init__(self, parent, style=wx_styles[style])
        self._align = (align == ALIGN_JUST)
        self._margin = margin
        self._spacing = spacing
        self._haswidgets = False
        wx_layout = layout[0] if isinstance(layout, tuple) else layout
        try:
            sizer_class, sizer_args = wx_layouts[wx_layout]
        except KeyError:
            raise ValueError("Invalid layout value: {}".format(wx_layout))
        if isinstance(layout, tuple):
            self._is_grid = True
            wx_count = layout[1]
            sizer_args += (wx_count,) if wx_layout == LAYOUT_COLGRID else (wx_count, 0)
            sizer_args += (spacing, spacing) if spacing else (0, 0)
        print(sizer_class, sizer_args)
        self._sizer = sizer_class(*sizer_args)
        PPanelBase.__init__(self, manager, parent, align, layout, contents=contents,
                            style=style, margin=margin, spacing=spacing)
    
    def set_min_size(self, width, height):
        self._sizer.SetMinSize((width, height))
    
    def set_margin(self, margin):
        # Margin is dealt with differently in wxWidgets, see above
        pass
    
    def set_spacing(self, spacing):
        # Spacing is dealt with differently in wxWidgets, see above
        pass
    
    def add_widget(self, widget):
        # People say the wxWidgets layout model is simple and easy to use,
        # but comparing this cruft with the PySide/Qt code, I'm not so sure...
        
        if self._is_grid:
            args = (wx.SizerFlags(0),)
        
        else:
            if self._haswidgets:
                if self._spacing is not None:
                    self._sizer.AddSpacer(self._spacing)
            elif self._margin is not None:
                self._sizer.AddSpacer(self._margin)
            
            if getattr(widget, '_align', None):
                proportion = 1
            else:
                proportion = 0
            
            if getattr(widget, '_expand', None):
                flag = wx.EXPAND
            else:
                flag = 0
            
            exp = getattr(widget, '_expand_horiz', None)
            if self._sizer.GetOrientation() == wx.VERTICAL:
                horiz_pad = self._margin or 0
                vert_pad = None
                if exp:
                    flag |= wx.EXPAND
            else:
                horiz_pad = None
                vert_pad = self._margin or 0
                if exp:
                    proportion = 1
            
            if horiz_pad is not None:
                flag |= wx.LEFT | wx.RIGHT
                border = horiz_pad
            elif vert_pad is not None:
                flag |= wx.TOP | wx.BOTTOM
                border = vert_pad
            else:
                border = 0
            
            args = (proportion, flag, border)
        
        # At last we can actually do what we came for...
        self._sizer.Add(widget, *args)
        if not self._haswidgets:
            self._haswidgets = True
    
    def do_layout(self):
        if (not self._is_grid) and (self._margin is not None):
            self._sizer.AddSpacer(self._margin)
        self.SetSizerAndFit(self._sizer)
    
    def remove_widget(self, widget):
        self._sizer.Detach(widget)
        self.SetSizerAndFit(self._sizer)



#!/usr/bin/env python3
"""
Module COMBO -- UI Combo Box Widgets
Sub-Package UI.BASE of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

import collections

from plib.stdlib.builtins import inverted
from plib.stdlib.coll import baselist, SortMixin

from plib.ui.defs import *
from plib.ui.coll import BaseStringListWidget

from .app import PWidgetBase


class PComboBoxBase(PWidgetBase, BaseStringListWidget):
    """Combo box that looks like a list of strings.
    
    (Note: currently selection is limited to items added to the
    combo box programmatically; the user cannot edit in the edit
    control and add new items to the pick list.)
    """
    
    signals = (
        SIGNAL_SELECTED,
    )
    
    def __init__(self, manager, parent, items=None, value=None, starting_index=None,
                 geometry=None, font=None):
        
        PWidgetBase.__init__(self, manager, parent,
                             geometry=geometry, font=font)
        BaseStringListWidget.__init__(self, items, value, starting_index)


class PNumComboBoxBase(PComboBoxBase):
    """Combo box that maps strings to integer values.
    """
    
    def __init__(self, manager, parent, items=None, value=None, starting_index=None,
                 geometry=None, font=None):
        
        self.text_map = dict(items)  # value -> text
        self.value_map = inverted(self.text_map)  # text -> value
        str_items = [item[1] for item in items]
        PComboBoxBase.__init__(self, manager, parent, items=str_items, starting_index=starting_index,
                               geometry=geometry, font=font)
        if value:
            self.set_current_value(value)  # value will be an int
    
    def current_value(self):
        return self.value_map[self.current_text()]
    
    get_current_value = current_value
    
    def set_current_value(self, value):
        self.set_current_text(self.text_map[value])


class PSortedComboBoxBase(SortMixin, PComboBoxBase):
    """Combo box that automatically sorts its items.
    """
    
    def __init__(self, manager, parent, items=None, value=None, starting_index=None,
                 geometry=None, font=None, key=None):
        
        PComboBoxBase.__init__(self, manager, parent, geometry=geometry, font=font)  # don't pass items here
        self._init_seq(items, key)  # sort items and add them here
        self.complete_init(items, value, starting_index)  # have to do this after data is added



#!/usr/bin/env python3
"""
Module QT5.LABEL -- Python Qt 5 Text Label Widgets
Sub-Package UI.TOOLKITS.QT5 of Package PLIB3 -- Python UI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the Qt 5 UI objects for text label widgets.
"""

from PyQt5 import QtWidgets as qt

from plib.ui.base.label import PTextLabelBase

from .app import PQtWidgetMeta, PQtWidgetBase


class PTextLabel(qt.QLabel, PQtWidgetBase, PTextLabelBase,
                 metaclass=PQtWidgetMeta):
    
    fix_width_on_resize = True
    
    widget_class = qt.QLabel
    
    def __init__(self, manager, parent, text=None,
                 geometry=None, font=None):
        
        qt.QLabel.__init__(self, parent)
        PTextLabelBase.__init__(self, manager, parent, text,
                                geometry=geometry, font=font)
    
    fn_get_text = 'text'
    fn_set_text = 'setText'



#!/usr/bin/env python3
"""
Module PYSIDE2.PAGEWIDGET -- Python PySide 2 Page Widget
Sub-Package UI.TOOLKITS.PYSIDE2 of Package PLIB3 -- Python UI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the PySide 2 UI objects for the page widget.
"""

from PySide2 import QtWidgets as qt

from plib.ui.defs import *
from plib.ui.base.pagewidget import PPageWidgetBase

from .app import PQtSequenceWidget


class PPageWidget(PQtSequenceWidget, qt.QStackedWidget, PPageWidgetBase):
    
    def __init__(self, manager, parent, pages=None, link_to=None):
        qt.QStackedWidget.__init__(self, parent)
        PPageWidgetBase.__init__(self, manager, parent, pages=pages, link_to=link_to)
    
    def count(self, value):
        # Method name collision, we want it to be the Python sequence
        # count method here.
        return PPageWidgetBase.count(self, value)
    
    def page_count(self):
        # Let this method access the Qt page widget count method.
        return qt.QStackedWidget.count(self)
    
    def page_at(self, index):
        return self.widget(index)
    
    def add_page(self, index, widget):
        self.insertWidget(index, widget)
    
    def del_page(self, index):
        self.removeWidget(index)
    
    def current_index(self):
        return self.currentIndex()
    
    def set_current_index(self, index):
        self.setCurrentIndex(index)



#!/usr/bin/env python3
"""
Module QT5.HTML -- Python Qt 5 Html Display Widgets
Sub-Package UI.TOOLKITS.QT5 of Package PLIB3 -- Python UI Toolkits
Copyright (C) 2008-2024 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the Qt 5 UI objects for text display widgets.
"""

from PyQt5 import QtWidgets as qt

from plib.ui.base.html import PHtmlDisplayBase

from .app import PQtWidgetMeta, PQtWidgetBase


class PHtmlDisplay(qt.QTextEdit, PQtWidgetBase, PHtmlDisplayBase,
                   metaclass=PQtWidgetMeta):
    
    fix_width_on_resize = True
    fix_height_on_resize = True
    
    widget_class = qt.QTextEdit
    
    def __init__(self, manager, parent, html=None,
                 scrolling=False, font=None):
        
        qt.QTextEdit.__init__(self, parent)
        self.setReadOnly(True)
        self.setSizePolicy(qt.QSizePolicy.MinimumExpanding, qt.QSizePolicy.MinimumExpanding)
        PHtmlDisplayBase.__init__(self, manager, parent, html, scrolling=scrolling, font=font)
    
    def setup_scrolling(self, scrolling):
        if scrolling:
            self.setLineWrapMode(qt.QTextEdit.NoWrap)
    
    def get_html(self):
        return self.html()
    
    def set_html(self, html):
        self.setHtml(html)



#!/usr/bin/env python3
"""
Sub-Package UI of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

__version__ = "0.15.10"



#!/usr/bin/env python3
"""
Module QT5.CONTAINER-- Python Qt 5 Container Objects
Sub-Package UI.TOOLKITS.PYSIDE2 of Package PLIB3 -- Python UI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the Qt 5 UI objects for the panel widgets.
"""

from PyQt5 import QtWidgets as qt

from plib.ui.base.container import PContainerBase

from .app import PQtWidgetBase


class PContainer(qt.QFrame, PQtWidgetBase, PContainerBase):
    
    def __init__(self, manager, parent):
        qt.QFrame.__init__(self, parent)
        self.setFrameStyle(qt.QFrame.NoFrame | qt.QFrame.Plain)
        self.setSizePolicy(qt.QSizePolicy.MinimumExpanding, qt.QSizePolicy.MinimumExpanding)
        self._playout = qt.QVBoxLayout()
        PContainerBase.__init__(self, manager, parent)
    
    def add_widget(self, widget):
        self._playout.addWidget(widget)
    
    def do_layout(self):
        #self._playout.addStretch(1)
        self.setLayout(self._playout)



#!/usr/bin/env python3
"""
Module COMMON -- Python UI Common Global Objects
Sub-Package UI of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

import sys
import os
import functools
import pkgutil

from plib.ui.defs import *
from plib.ui.toolkits import get_toolkit


# action dictionary
action_map = {
    ACTION_FILE_NEW: ["file_new", "&New", "Create new document"],
    ACTION_FILE_OPEN: ["file_open", "&Open", "Open document"],
    ACTION_FILE_SAVE: ["file_save", "&Save", "Save current document"],
    ACTION_FILE_SAVEAS: ["file_saveas", "Save &As", "Save current document with different filename"],
    ACTION_FILE_CLOSE: ["file_close", "&Close", "Close current document"],
    ACTION_VIEW: ["view", "&View", "View item"],
    ACTION_EDIT: ["edit", "&Edit", "Edit item"],
    ACTION_EDIT_UNDO: ["edit_undo", "&Undo", "Undo last operation"],
    ACTION_EDIT_REDO: ["edit_redo", "&Redo", "Redo last operation"],
    ACTION_EDIT_CUT: ["edit_cut", "Cu&t", "Cut selection to clipboard"],
    ACTION_EDIT_COPY: ["edit_copy", "&Copy", "Copy selection to clipboard"],
    ACTION_EDIT_PASTE: ["edit_paste", "&Paste", "Paste from clipboard"],
    ACTION_EDIT_DELETE: ["edit_delete", "&Delete", "Delete selection"],
    ACTION_EDIT_SELECTALL: ["edit_selectall", "Select &All", "Select all"],
    ACTION_EDIT_SELECTNONE: ["edit_selectnone", "Clear &Selection", "Clear selection"],
    ACTION_EDIT_OVERWRITE: ["edit_overwrite", "Toggle &Overwrite", "Toggle overwrite mode"],
    ACTION_EDIT_CLEAR: ["edit_clear", "Clear &Edit", "Clear document"],
    ACTION_REFRESH: ["refresh", "Re&fresh", "Refresh view"],
    ACTION_ADD: ["add", "A&dd", "Add item"],
    ACTION_REMOVE: ["remove", "Re&move", "Remove item"],
    ACTION_APPLY: ["apply", "&Apply", "Apply changes"],
    ACTION_COMMIT: ["commit", "Co&mmit", "Commit transaction"],
    ACTION_ROLLBACK: ["rollback", "Rollbac&k", "Roll back transaction"],
    ACTION_OK: ["ok", "&Ok", "Ok"],
    ACTION_CANCEL: ["cancel", "&Cancel", "Cancel"],
    ACTION_PREFS: ["prefs", "&Preferences...", "Edit preferences"],
    ACTION_ABOUT: ["about", "A&bout...", "About this program"],
    ACTION_ABOUT_TOOLKIT: [
        "about_toolkit",
        "Abou&t {}...".format(get_toolkit().strip('2').strip('5').capitalize()),
        "About this UI toolkit"
    ],
    ACTION_EXIT: ["exit", "E&xit", "Exit program"]
}

# action key list (needed to ensure proper ordering of actions,
# since the dictionary keys won't necessarily be ordered)

action_key_list = sorted(action_map.keys())


# Allow adding custom actions; must be called before any
# widgets are created

def add_action(key, name, caption, description, icon):
    if key in action_map:
        raise RuntimeError("Key {} is already in use".format(key))
    action_map[key] = [name, caption, description, icon]
    action_key_list.append(key)
    action_key_list.sort()


# Allow modifying existing actions

def modify_action(key, name=None, caption=None, description=None, icon=None):
    entry = action_map.get(key)
    if not entry:
        raise RuntimeError("Invalid action key: {}".format(key))
    items = (name, caption, description, icon)
    add_index = len(entry)
    for index, item in enumerate(items[:add_index]):
        if item:
            entry[index] = item
    add_items = items[add_index:]
    if add_items:
        entry[add_index:] = add_items


# utility functions to get fully qualified pixmap path name


def pxdata(name,
           pxfmt="images/{}.png".format, fmt="{}-{}".format,
           get_data=functools.partial(pkgutil.get_data, 'plib.ui')):
    
    for suffix in (get_toolkit(), sys.platform, os.name, None):
        try:
            data = get_data(pxfmt(fmt(name, suffix) if suffix else name))
        except OSError:
            data = None
        if data:
            return data
    if os.path.isfile(name):
        with open(name, 'rb') as f:
            data = f.read()
        return data


# Convenience functions for action properties

def action_name(action):
    return action_map[action][0]


def action_caption(action):
    return action_map[action][1]


def action_description(action):
    return action_map[action][2]


def action_icondata(action):
    props = action_map[action]
    # This hack allows the icon data key to be the name or be set separately
    return pxdata(props[3 if len(props) > 3 and props[3] else 0])


# Process font arguments to ensure all are filled in

def font_args(widget, font_name, font_size, bold, italic):
    if font_name is None:
        font_name = widget.get_font_name()
    if font_size is None:
        font_size = widget.get_font_size()
    if bold is None:
        bold = widget.get_font_bold()
    if italic is None:
        italic = widget.get_font_italic()
    return font_name, font_size, bold, italic


# Convenience operations for font specs

def set_name(font, name=None):
    post = font[1:]
    return (name,) + post


def set_size(font, size=None):
    pre = font[:1]
    pre += (None,) * (1 - len(pre))
    post = font[2:]
    return pre + (size,) + post


def set_bold(font, bold=True):
    pre = font[:2]
    pre += (None,) * (2 - len(pre))
    post = font[3:]
    return pre + (bold,) + post


def set_italic(font, italic=True):
    pre = font[:3]
    pre += (None,) * (3 - len(pre))
    return pre + (italic,)



#!/usr/bin/env python3
"""
Sub-Package UI.WIDGETS of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

from plib.stdlib.classtools import recurselist

from plib.ui.defs import *
from plib.ui.imp import get_toolkit_object
from plib.ui.common import action_name


# Helper function to find and connect signals to handlers

def setup_signal_handlers(manager, obj, name,
                          event_map=WIDGET_EVENT_MAP):
    
    for signal in recurselist(type(obj), 'signals'):
        event = event_map[signal]
        method = "on_{}_{}".format(name, event) if event else "on_{}".format(name)
        target = getattr(manager, method, None)
        if target:
            obj.setup_notify(signal, target)


# Basic widget construction function; this will end up getting called
# for each widget spec with the obvious arguments. The widget will
# automagically find its event handlers on the manager when it is constructed

def widget_from_spec(manager, parent, spec,
                     attr_prefixes=WIDGET_ATTR_PREFIXES, auto_prefixes=WIDGET_AUTO_PREFIXES):
    
    # We do the test of the spec this way to avoid having to import
    # the module with PWidgetManager in it; this avoids circular imports
    # and also avoids having to import that code if it is not used
    if isinstance(spec, tuple):
        widget_manager_klass = None
        widget_manager = manager
    else:  # assume it's a subclass of PWidgetManager
        widget_manager_klass = spec
        widget_manager = widget_manager_klass(manager)
        setattr(manager, widget_manager.attrname, widget_manager)
        # This must be a tuple (no point in it being another PWidgetManager subclass)
        spec = widget_manager.main_widget
    
    modname, klassname, name, args, kwargs = spec
    klass = get_toolkit_object(modname, klassname)
    attrname = (
        name
        if any(name.startswith(prefix) for prefix in auto_prefixes) else
        "{}_{}".format(attr_prefixes.get(klassname, modname), name)
    )
    widget = klass(widget_manager, parent, *args, **kwargs)
    setattr(widget_manager, attrname, widget)
    setup_signal_handlers(widget_manager, widget, name)
    return widget


# Helper function for widgets that can have contents as list of specs

def widgets_from_contents(manager, parent, contents, callback=None):
    widgets = []
    for spec in contents:
        widget = widget_from_spec(manager, parent, spec)
        if callback:
            callback(widget)
        widgets.append(widget)
    return widgets


# Widget spec convenience functions

def auto_name(prefix, num):
    prefix = '{}_'.format(prefix)
    num[0] += 1
    return '{}{}'.format(prefix, num[0])


## Widget spec functions ##


# Buttons have a caption and optional icon
# Buttons fire SIGNAL_CLICKED

def button(name, caption, icon=None, font=None):
    args = (caption,)
    kwargs = dict(
        icon=icon,
        geometry=(None, None, BUTTON_WIDTH, None),
        font=font,
    )
    return ('button', 'PButton', name, args, kwargs)


# Action buttons take their caption and icon from a menu/toolbar action
# Action buttons fire SIGNAL_ACTIVATED when clicked

def action_button(action, font=None):
    name = action_name(action)
    args = (action,)
    kwargs = dict(
        geometry=(None, None, BUTTON_WIDTH, None),
        font=font,
    )
    return ('button', 'PActionButton', name, args, kwargs)


# Button groups are groups of toggle or radio buttons that are mutually exclusive
# Button groups fire SIGNAL_BUTTONSELECTED when the selected button is changed

def base_buttongroup(klassname, name, align, layout, items=None, value=None, starting_index=None, font=None):
    args = (align, layout, items)
    kwargs = dict(starting_index=starting_index, font=font)
    return ('group', klassname, name, args, kwargs)


def buttongroup(name, align, layout, items=None, value=None, starting_index=None, font=None):
    return base_buttongroup('PButtonGroup', name, align, layout, items, value, starting_index, font)


def radiogroup(name, align, layout, items=None, value=None, starting_index=None, font=None):
    return base_buttongroup('PRadioGroup', name, align, layout, items, value, starting_index, font)


# Check boxes have a label and checked state
# Check boxes fire SIGNAL_TOGGLED

def checkbox(name, label, checked=None, font=None):
    args = (label, checked)
    kwargs = dict(
        font=font,
    )
    return ('button', 'PCheckBox', name, args, kwargs)


def base_combo(klassname, name, items, value, **kwds):
    args = (items,)
    kwargs = dict(
        value=value,
    )
    kwargs.update(kwds)
    return ('combo', klassname, name, args, kwargs)


# Combo boxes have a string value and a drop-down list of choices
# Combo boxes fire SIGNAL_SELECTED

def combo(name, items=None, value=None, geometry=None, font=None):
    return base_combo('PComboBox', name, items, value, geometry=geometry, font=font)


# Numeric combo boxes present their values as ints instead of string

def num_combo(name, items=None, value=None, geometry=None, font=None):
    return base_combo('PNumComboBox', name, items, value, geometry=geometry, font=font)


# Sorted combo boxes sort their strings in the drop-down list

def sorted_combo(name, items, value=None, key=None, geometry=None, font=None):
    return base_combo('PSortedComboBox', name, items, value, geometry=geometry, font=font, key=key)


def base_edit(klassname, name, args, geometry, font, expand=True):
    kwargs = dict(
        geometry=geometry,
        font=font,
        expand=expand,
    )
    return ('editctrl', klassname, name, args, kwargs)


# Edit boxes are single-line string editors and can be initialized with a value;
# they default to automatically expanding horizontally to fill available space
# Edit boxes fire: SIGNAL_EDITCHANGED when their content is changed
#                  SIGNAL_ENTER when Enter is pressed (normally used to trigger an action)

def edit(name, text="", geometry=None, font=None, expand=True):
    #geometry = geometry or (None, None, EDIT_WIDTH, None)
    args = (text,)
    return base_edit('PEditBox', name, args, geometry, font, expand=expand)


# Numeric edit boxes present their value as an int instead of a string

def num_edit(name, value=0, geometry=None, font=None, expand=True):
    args = (value,)
    #geometry = geometry or (None, None, NUM_EDIT_WIDTH, None)
    return base_edit('PNumEditBox', name, args, geometry, font, expand=expand)


# Numeric edit with spinner

def spin_edit(name, value=0, min=0, max=100, step=1, geometry=None, font=None, expand=True):
    args = (value, min, max, step)
    return base_edit('PSpinEditBox', name, args, geometry, font, expand=expand)


# Group boxes organize child controls; they automatically expand to
# fill available space; contents must be an iterable of widget specs

def groupbox(caption, contents, name=None, font=None,
             num=[0]):
    
    name = name or auto_name('groupbox', num)
    args = (caption,)
    kwargs = dict(
        contents=contents,
        margin=TAB_MARGIN,
        spacing=PANEL_SPACING,
        font=font,
    )
    return ('groupbox', 'PGroupBox', name, args, kwargs)


# A label displays its caption string

def label(name, caption, geometry=None, font=None):
    #geometry = geometry or (None, None, LABEL_WIDTH, None)
    args = (caption,)
    kwargs = dict(
        geometry=geometry,
        font=font,
    )
    return ('label', 'PTextLabel', name, args, kwargs)


def base_listbox(klassname, name, items, value, **kwargs):
    args = (items, value)
    return ('listbox', klassname, name, args, kwargs)


# List boxes display a list of strings
# List boxes fire SIGNAL_LISTBOXSELECTED

def listbox(name, items=None, value=None, geometry=None, font=None):
    #geometry = geometry or (None, None, LIST_WIDTH, LIST_HEIGHT)
    return base_listbox('PListBox', name, items, value, geometry=geometry, font=font)


# Sorted list boxes sort their displayed list of strings

def sorted_listbox(name, items=None, value=None, key=None, geometry=None, font=None):
    #geometry = geometry or (None, None, LIST_WIDTH, LIST_HEIGHT)
    return base_listbox('PSortedListBox', name, items, value, key=key, geometry=geometry, font=font)


def base_listview(klassname, name, labels, **kwds):
    kwargs = dict(
        labels=labels,
    )
    kwargs.update(kwds)
    return ('listview', klassname, name, (), kwargs)


# List views display rows each with one or more columns of strings, plus a header;
# they automatically expand to fill available space
# List views appear in code as sequences of tuples of strings
# List views fire SIGNAL_LISTVIEWSELECTED

def listview(name, labels=None, data=None, font=None, header_font=None):
    return base_listview('PListView', name, labels=labels, data=data,
                         font=font, header_font=header_font)


# Sorted list views sort their displayed rows, treating each row as a tuple of strings

def sorted_listview(name, labels=None, data=None, key=None, font=None, header_font=None):
    return base_listview('PSortedListView', name, labels=labels, data=data, key=key,
                         font=font, header_font=header_font)


# Tree views display a tree of items, each with one or more columns of strings, plus a header;
# they automatically expand to fill available space
# Tree views appear in code as sequences of 2-tuples (cols, children), where cols is a tuple
# of strings and children is another sequence of 2-tuples (cols, children)
# Tree views fire SIGNAL_LISTVIEWSELECTED

def treeview(name, labels=None, data=None, auto_expand=False, font=None, header_font=None):
    return base_listview('PTreeView', name, labels=labels, data=data, auto_expand=auto_expand,
                         font=font, header_font=header_font)


# Edit controls (memos) are multi-line scrolling plain text editors;
# they automatically expand to fill available space
# Edit controls fire: SIGNAL_TEXTCHANGED when their text is changed
#                     SIGNAL_TEXTSTATECHANGED when their text selection state is changed

def memo(name, scrolling=True, font=None):
    kwargs = dict(
        scrolling=scrolling,
        font=font,
    )
    return ('editctrl', 'PEditControl', name, (), kwargs)


# Padding is a blank panel that automatically expands to fill available space;
# it can be used to help align other widgets in layouts

def padding(name=None,
            num=[0]):
    
    name = name or auto_name('padding', num)
    return ('form', 'PPanel', name, (), {})


# The contents for all panels must be an iterable of widget specs

def base_panel(align, layout, contents, style=PANEL_PLAIN, margin=-1, spacing=-1, name=None,
               num=[0]):
    
    name = name or auto_name('panel', num)
    args = (align, layout)
    kwargs = dict(
        contents=contents,
        style=style,
        margin=margin,
        spacing=spacing,
    )
    return ('form', 'PPanel', name, args, kwargs)


# Frames are top-level layouts typically used as the main widget in an application to
# organize the layout of multiple sub-widgets; they automatically expand to fill
# available space

def frame(align, layout, contents, style=PANEL_PLAIN, name=None):
    return base_panel(align, layout, contents, style=style, margin=FRAME_MARGIN, spacing=PANEL_SPACING, name=name)


# Panels are layout widgets typically used as sub-widgets to frames in complex
# layouts; they automatically expand to fill available space

def panel(align, layout, contents, style=PANEL_PLAIN, name=None):
    return base_panel(align, layout, contents, style=style, margin=PANEL_MARGIN, spacing=PANEL_SPACING, name=name)


# Tabs are panels specialized to organize multiple widgets in a tab widget's tab;
# they automatically expand to fill available space

def tab(align, layout, contents, style=PANEL_PLAIN, name=None):
    return base_panel(align, layout, contents, style=style, margin=TAB_MARGIN, spacing=PANEL_SPACING, name=name)


# Boxes are simple containers typically used at the bottom level of layouts;
# they automatically expand to fill available space

def box(align, layout, contents, style=PANEL_PLAIN, name=None):
    return base_panel(align, layout, contents, style=style, margin=BOX_MARGIN, spacing=PANEL_SPACING, name=name)


# Tables are grid-style widgets with multiple rows and columns plus a header;
# they automatically expand to fill available space

def table(name, labels=None, data=None, font=None, header_font=None):
    kwargs = dict(
        labels=labels,
        data=data,
        font=font,
        header_font=header_font,
    )
    return ('table', 'PTable', name, (), kwargs)


# Tab widgets organize complex layouts in multiple tabs; they automatically expand
# to fill available space; tabs must be an iterable of widget specs

def tabwidget(name, tabs, font=None):
    kwargs = dict(
        tabs=tabs,
        font=font,
    )
    return ('tabwidget', 'PTabWidget', name, (), kwargs)


# Page widgets organize complex layouts in multiple pages; they automatically expand
# to fill available space; pages must be an iterable of widget specs; link_to gives
# the name of another widget that will be used to select the current page

def pagewidget(name, pages, link_to=None):
    kwargs = dict(
        pages=pages,
        link_to=link_to,
    )
    return ('pagewidget', 'PPageWidget', name, (), kwargs)


# Text displays are scrolling widgets that show multiple lines of plain text;
# they automatically expand to fill available space

def text(name, text="", scrolling=True, font=None):
    args = (text,)
    kwargs = dict(
        scrolling=scrolling,
        font=font,
    )
    return ('display', 'PTextDisplay', name, args, kwargs)


# Html displays are scrolling widgets that show html; they automatically
# expand to fill available space

def html(name, html="", scrolling=True, font=None):
    args = (html,)
    kwargs = dict(
        scrolling=scrolling,
        font=font,
    )
    return ('html', 'PHtmlDisplay', name, args, kwargs)


# The scrolling wrapper puts a scrolling window over a child widget
# specified by child_spec

def scrolling(child_spec):
    name = child_spec[2]
    args = (child_spec,)
    kwargs = dict()
    return ('scroller', 'PScrollingWrapper', name, args, kwargs)


# Containers are wrappers around widgets that are not directly supported
# by PLIB.UI; the add_child method is used to add the custom widget as
# a single child widget of the container

def container(name):
    return ('container', 'PContainer', name, (), {})


# Convenience function for labeled controls

def labeled(caption, control, layout=LAYOUT_VERTICAL, align=ALIGN_JUST, label_bold=True, label_font=None):
    name = control[2]  # control will be a widget spec tuple
    label_align = (ALIGN_TOP if layout == LAYOUT_VERTICAL else ALIGN_LEFT)
    label_name = '{}_label'.format(name)
    labelbox_name = '{}_labelbox'.format(name)
    box_name = '{}_box'.format(name)
    return box(align, layout, [
        box(label_align, LAYOUT_HORIZONTAL, [
            # Label name will be prefixed by 'label' in widget_from_spec
            label(label_name, caption, font=(label_font, None, label_bold)),
        ], name=labelbox_name),
        control
    ], name=box_name)

# Label boxes work like group boxes but allow a bold label without making
# every widget in the box bold also

def labelbox(caption, contents, layout=LAYOUT_VERTICAL, align=ALIGN_JUST, style=PANEL_PLAIN, label_bold=True, label_font=None, name=None,
             num=[0]):
    
    name = name or auto_name('labelbox', num)
    label_name = '{}_label'.format(name)
    box_name = '{}_box'.format(name)
    panel_name = '{}_panel'.format(name)
    contents = panel(ALIGN_JUST, layout, contents, style=style, name=panel_name)
    return box(align, LAYOUT_VERTICAL, [
        label(label_name, caption, font=(label_font, None, label_bold)),
        contents
    ], name=box_name)



#!/usr/bin/env python3
"""
Module IMP -- Toolkit Import Support
Sub-Package UI of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

Defines support functions for importing toolkit objects.
"""

from plib.stdlib.imp import import_from_module

from plib.ui.toolkits import get_toolkit


# Toolkit-aware object retrieval function

def get_toolkit_object(modname, name):
    if ("." not in modname) and (modname != '__main__'):
        toolkit = get_toolkit()
        modname = "plib.ui.toolkits.{}.{}".format(toolkit, modname)
    return import_from_module(modname, name)



#!/usr/bin/env python3
"""
Module PAGEWIDGET -- UI Page Widget
Sub-Package UI.BASE of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

from plib.ui.defs import *
from plib.ui.coll import BaseListWidget
from plib.ui.widgets import widget_from_spec

from .app import PWidgetBase


class PPageWidgetBase(PWidgetBase, BaseListWidget):
    """Page widget as a Python list of widgets.
    """
    
    signals = (
        SIGNAL_TABSELECTED,
    )
    
    def __init__(self, manager, parent, pages=None, link_to=None):
        PWidgetBase.__init__(self, manager, parent)
        BaseListWidget.__init__(self)
        if pages:
            self.create_pages(pages)
        self.link_to = link_to
    
    def link_to_widget(self, widget):
        signal = widget.signals[0]  # assume this is the selected signal that gives the new index
        widget.setup_notify(signal, self.on_link_selected)
    
    def setup_link_to(self):
        if not self.link_to:
            return
        widget = getattr(self.manager, self.link_to)
        self.link_to_widget(widget)
        self.link_to = None  # so repeated calls won't do anything
    
    def on_link_selected(self, index):
        # Need to filter signals here since it is possible for the selector
        # to send an out of range index if it is mutated; we don't raise an
        # exception because mutations should be due to other code, so the
        # code writer just needs to be aware of the consequences
        if (index < 0) or (index > (len(self) - 1)):
            return
        self.set_current_index(index)
    
    def setup_signals(self):
        self.setup_link_to()
    
    def create_pages(self, pages):
        """Create pages from sequence.
        
        Assumes that the sequence contains a widget spec for each page.
        """
        self.extend(widget_from_spec(self.manager, self, spec) for spec in pages)
    
    def append_and_focus(self, page):
        """Add new page and set focus to it.
        
        Assumes that page is a widget spec.
        """
        self.append(page)
        self.set_current_index(len(self) - 1)
    
    def __len__(self):
        return self.page_count()
    
    def _indexlen(self):
        return self.page_count()
    
    def page_count(self):
        """ Placeholder for derived classes to implement. """
        raise NotImplementedError
    
    def page_at(self, index):
        """ Placeholder for derived classes to implement. """
        raise NotImplementedError
    
    def add_page(self, index, widget):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def del_page(self, index):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def _get_data(self, index):
        return self.page_at(index)
    
    def _set_data(self, index, value):
        # This gets a little complicated because we have to delete the
        # old page at this index and insert a new one.
        self._del_data(index)
        self._add_data(index, value)
    
    def _add_data(self, index, value):
        self.add_page(index, value)
    
    def _del_data(self, index):
        self.del_page(index)
    
    def current_widget(self):
        return self[self.current_index()]



#!/usr/bin/env python3
"""
Module OUTPUT-- UI Output Classes
Sub-Package UI of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""


class PTextOutput(object):
    """File-like object to redirect output to text control.
    
    Note that 'file-like' is used very loosely; this object only
    implements the minimum necessary to make a PTextDisplay
    look like a bare bones file-like object. In particular,
    iterator functionality is *not* implemented.
    """
    
    def __init__(self, control):
        self.control = control  # should normally be a PTextDisplay
    
    def close(self):
        pass  # nothing to do here
    
    @property
    def filesize(self):
        return len(self.control.get_text())
    
    def truncate(self, size=0):
        if size == 0:
            s = ""
        else:
            s = self.control.get_text()[:size]
        self.control.set_text(s)
        self.flush()
    
    def read(self):
        return self.control.get_text()
    
    def write(self, data):
        self.control.add_text(data)
    
    def flush(self):
        self.control.update_widget()
    
    def clear(self):
        self.control.clear_text()



#!/usr/bin/env python3
"""
Module MAINWIN -- UI Main Window Classes
Sub-Package UI.BASE of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

Defines the classes for fully functional decorated GUI main
windows. This module is kept separate from the ``app`` module
so that applications that do not require a full main window
but just a "bare" top window do not have to import any code
that is only used with a full main window.
"""

from plib.ui.defs import *
from plib.ui.common import *
from plib.ui.widgets import setup_signal_handlers, widget_from_spec

from .app import PSignalBase, PActionMixin, PTopWindowBase


class PUIBase(object):
    """Base class for UI menus and toolbars.
    """
    
    def __init__(self, main_window):
        self.main_window = main_window


class PMenuBase(PUIBase):
    """Base class for UI menu.
    """
    
    popup_class = None
    
    popups = None
    
    def store_popup(self, title, popup):
        if self.popups is None:
            self.popups = {}
        self.popups[title] = popup
    
    def do_add_popup(self, title, popup):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def add_popup(self, title):
        popup = self.popup_class(self.main_window)
        self.do_add_popup(title, popup)
        self.store_popup(title, popup)
        return popup
    
    def get_popup(self, title):
        return self.popups.get(title) if self.popups else None
    
    def add_popup_action(self, action, popup):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError


class PToolBarBase(PUIBase):
    """Base class for UI toolbar.
    """
    
    def add_action(self, action):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def add_separator(self):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError


class PStatusLabelBase(object):
    """Base class for status bar labels.
    """
    
    actual_widget = None
    
    def __init__(self, manager, parent, caption="", style=PANEL_PLAIN):
        self._manager = manager
        self._parent = parent
        self._caption = caption
        self._style = style
        self.init_status_label()
    
    def init_status_label(self):
        raise NotImplementedError
    
    fn_get_text = None
    fn_set_text = None
    
    def get_text(self):
        return getattr(self, self.fn_get_text)()
    
    def set_text(self, value):
        getattr(self, self.fn_set_text)(value)
    
    caption = property(get_text, set_text)


class PStatusBarBase(object):
    """Base class for UI status bar.
    
    A standard status bar with a text area on the left and an
    area for custom widgets on the right.
    """
    
    text_area_class = None
    
    def __init__(self, main_window):
        self.main_window = main_window
        self.create_text_area()
        self.create_widgets()
    
    def add_widget(self, widget, expand=False, custom=True):
        """Placeholder for derived classes to implement
        """
        raise NotImplementedError
    
    def create_widgets(self,
                       kwargs={}):
        
        app = self.main_window.app
        labels = app.status_labels
        if labels is not None:
            for label_spec in labels:
                name, *args = label_spec
                label = widget_from_spec(app, self, ('mainwin', 'PStatusLabel', name, args, kwargs))
                self.add_widget(label.actual_widget)
    
    def create_text_area(self):
        self.text_area = None
        if self.text_area_class is not None:
            self.text_area = self.text_area_class(self.main_window.app, self)
            # This should be the only widget added with expand=True and custom=False
            self.add_widget(self.text_area, True, False)
    
    def get_text(self):
        return self.text_area.get_text()
    
    def set_text(self, value):
        self.text_area.set_text(value)


class PActionBase(PActionMixin, PSignalBase):
    """Base class for UI action objects.
    """
    
    signals = (
        SIGNAL_ACTIVATED,
    )
    
    def __init__(self, key, main_window):
        self.key = key
        self.main_window = main_window
    
    def enable(self, enabled):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError


class PMainWindowBase(PTopWindowBase):
    """Base class for 'main window' widgets.
    
    The main window is a fancier version of the top window, which
    includes all the frills that PTopWindow leaves out.
    """
    
    menu_class = None
    toolbar_class = None
    statusbar_class = None
    action_class = None
    
    actions = None
    
    def before_client(self):
        self.menu = self.create_menu() if self.app.include_menu else None
        self.toolbar = self.create_toolbar() if self.app.include_toolbar else None
        self.statusbar = self.create_statusbar() if self.app.include_statusbar else None
        
        self.actions = self.create_actions()
    
    def create_menu(self):
        """Create the main window's menu.
        """
        return self.menu_class(self)
    
    def create_toolbar(self):
        """Create the main window's toolbar.
        """
        return self.toolbar_class(self)
    
    def create_statusbar(self):
        """Create the main window's status bar.
        """
        return self.statusbar_class(self)
    
    def action_exists(self, key):
        """Return True if action key is in our list of actions.
        """
        return (key in self.app.action_flags)
    
    def create_action(self, key):
        """Return an action object for key.
        """
        return self.action_class(key, self)
    
    def get_action(self, actions, key):
        action = actions.get(key)
        if not action:
            actions[key] = action = self.create_action(key)
            setup_signal_handlers(self.app, action, action_name(key))
        return action
    
    def create_actions(self):
        """Create actions and link them to menu and toolbar items.
        """
        
        actions = {}
        
        for title, action_list in (self.app.menu_actions or ()):
            popup = self.menu.add_popup(title)
            for key in action_list:
                action = self.get_action(actions, key)
                self.menu.add_popup_action(action, popup)
        
        toolbar_started = False
        for action_list in (self.app.toolbar_actions or ()):
            if toolbar_started:
                self.toolbar.add_separator()
            else:
                toolbar_started = True
            for key in action_list:
                action = self.get_action(actions, key)
                self.toolbar.add_action(action)
        
        return actions
    
    def update_action(self, key, flag):
        """Update action enable state based on flag.
        """
        
        if self.action_exists(key):
            self.actions[key].enable(flag)
    
    def menu_height(self):
        return self.menu.preferred_height()
    
    def toolbar_height(self):
        return self.toolbar.preferred_height()
    
    def statusbar_height(self):
        return self.statusbar.preferred_height()
    
    def get_client_size(self):
        h = self.client_widget.preferred_height()
        if self.menu is not None:
            h += self.menu_height()
        if self.toolbar is not None:
            h += self.toolbar_height()
        if self.statusbar is not None:
            h += self.statusbar_height()
        w = 0
        for widget in (self.client_widget, self.toolbar):
            w = max(w, widget.preferred_width())
        return (w, h)



#!/usr/bin/env python3
"""
SCRIPS-EDIT.PY
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

Editor for scrips.dat file used to keep track of
prescription refills.
"""

import sys
import os
import datetime
import email.charset

from plib.ui import __version__
from plib.ui.defs import *
from plib.ui import common
from plib.ui.app import PApplication
from plib.ui.prefs import PPreferences
from plib.ui.widgets import *


def _load_example_module():
    from plib.stdlib.imp import import_from_path
    from plib.stdlib.postinstall import get_share_dir
    
    return import_from_path(
        os.path.join(
            get_share_dir('plib3.stdlib', 'plib.stdlib'),
            'examples', 'scrips'
        ),
        'scrips3'
    )


scrips = _load_example_module()


# Add custom actions
ACTION_SUBMIT = ACTION_OK + 1
ACTION_REFILL = ACTION_REFRESH + 1

common.add_action(
    ACTION_SUBMIT,
    'submit', "&Submit", "Submit prescription for refill", "ok"
)

common.add_action(
    ACTION_REFILL,
    'refill', "&Refill", "Complete prescription refill", "refresh"
)

# Modify some action descriptions
common.modify_action(ACTION_FILE_SAVE, description="Save prescription list")
common.modify_action(ACTION_ADD, description="Add prescription to list")
common.modify_action(ACTION_REMOVE, description="Remove prescription from list")


# Last item in each tuple is character column width for text output

SCRIP_HEADINGS = [
    ("Drug", 150, 16),
    ("Rx", 100, 12),
    ("Last Filled", 100, ALIGN_CENTER, 16),
    ("Days", 100, ALIGN_CENTER, 8),
    ("Refills Left", 100, ALIGN_CENTER, 16),
    ("Submitted", 100, 0)
]


def output_label(index, heading):
    return "{}{}".format(("#" if index == 0 else ""), heading[0]).ljust(heading[-1])


class ScripEditable(scrips.Scrip):
    
    init_name = "<Name>"
    init_rxnum = "<Rx#>"
    init_days = 30
    init_refills = 0
    init_submitted = False
    
    headings = SCRIP_HEADINGS
    
    def __init__(self, *tokens):
        if not tokens:
            # The default of today's date takes care of the filldate field
            # since there's no init class field for it
            tokens = [
                str(getattr(self, 'init_{}'.format(name), datetime.date.today()))
                for name in self._fields
            ]
        scrips.Scrip.__init__(self, *tokens)
    
    def submit(self):
        if self.refills < 1:
            return False
        # TODO: actually send the email here?
        self.submitted = True
        return True
    
    def refill(self):
        if not self.submitted:
            return False
        self.filldate = datetime.date.today()
        self.refills -= 1
        self.submitted = False
        return True
    
    def output_line(self):
        return "".join([
            str(self[col]).ljust(heading[-1])
            for col, heading in enumerate(self.headings)
        ])


class ScripListEditor(PApplication):
    
    about_data = {
        'name': "ScripsEdit",
        'version': "{} on Python {}".format(
            __version__,
            sys.version.split()[0]
        ),
        'description': "Prescription Editor",
        'copyright': "Copyright (C) 2008-2022 by Peter A. Donis",
        'license': "GNU General Public License (GPL) Version 2",
        'developer': "Peter Donis",
        'website': "http://www.peterdonis.net",
    }
    
    about_format = "{name} {version}\n\n{description}\n\n{copyright}\n{license}\n\nDeveloped by {developer}\n{website}"
    
    prefs_data = (scrips.inifile, SECTION_GROUPBOX, {
        "email": "E-Mail Fields",
        "email_fromaddr": "From",
        "email_toaddr": "To",
        "email_typestr": "MIME Type",
        "email_charsetstr": "Character Set",
        "email_serverstr": "Server Hostname",
        "email_portnum": "Server Port",
        "email_username": "User Name",
        "email_password": "Password",
        "headers": "E-Mail Headers",
        "headers_dict": "Python Dictionary",
        "pharmacy": "Pharmacy",
        "pharmacy_name": "Name",
    })
    
    prefs_dialog_enums = {
        "email_typestr": (
            "text/plain",
            "text/html",
        ),
        "email_charsetstr": tuple(sorted(
            email.charset.CHARSETS.keys()
        )),
        "email_portnum": (
            (25, "Plain SMTP"),
            (465, "Old SMTP/TLS"),
            (587, "SMTP/TLS"),
        ),
    }
    
    prefs_dialog_groups = {
        "email": (
            ("email_fromaddr", "email_toaddr"),
            ("email_typestr", "email_charsetstr"),
            ("email_serverstr", "email_portnum"),
            ("email_username", "email_password"),
        ),
    }
    
    prefs_dialog_width = 640
    
    menu_actions = [
        (MENU_FILE, (ACTION_FILE_SAVE, ACTION_EXIT,)),
        (MENU_ACTION, (ACTION_SUBMIT, ACTION_REFILL, ACTION_ADD, ACTION_REMOVE,)),
        (MENU_OPTIONS, (ACTION_PREFS,)),
        (MENU_HELP, (ACTION_ABOUT, ACTION_ABOUT_TOOLKIT,)),
    ]
    
    toolbar_actions = [
        (ACTION_FILE_SAVE,),
        (ACTION_SUBMIT, ACTION_REFILL, ACTION_ADD, ACTION_REMOVE,),
        (ACTION_PREFS, ACTION_ABOUT, ACTION_ABOUT_TOOLKIT, ACTION_EXIT,),
    ]
    
    main_title = "Prescription List Editor"
    main_iconfile = os.path.join(os.path.split(os.path.realpath(__file__))[0],
                                 "scrips.png")
    
    main_size = SIZE_CLIENTWRAP
    main_placement = MOVE_CENTER
    
    headings = SCRIP_HEADINGS
    
    table_font_size = 16 if sys.platform == 'darwin' else 12
    
    main_widget = table('scrips', [heading[:-1] for heading in headings],
                        font=("Arial", table_font_size),
                        header_font=("Arial", table_font_size, True))
    
    large_icons = True
    show_labels = True
    
    filename = scrips.scrips_dat_file()
    
    def after_create(self):
        self.prefs = PPreferences(self, "ScripsEdit Preferences")
        self.data = scrips.scriplist(ScripEditable)
        self.table_scrips.extend(self.data)
        for row in range(len(self.data)):
            self.set_colors(row)
        self.main_window.statusbar.set_text("Editing prescription info.")
        self.pending = False
        self.modified = False
        self.update_pending()
        self.update_size()
    
    def update_pending(self):
        self.pending = any(
            (self.data[row].due() and not self.data[row].submitted)
            for row in range(len(self.data))
        )
    
    def set_colors(self, row=None):
        if row is None:
            row = self.table_scrips.current_row()
        scrip = self.data[row]
        if scrip.due():
            if scrip.submitted:
                self.table_scrips.set_row_fgcolor(row, COLOR_BLACK)
            else:
                self.table_scrips.set_row_fgcolor(row, COLOR_RED)
            self.table_scrips.set_row_bkcolor(row, COLOR_YELLOW)
        else:
            self.table_scrips.set_row_fgcolor(row, COLOR_BLACK)
            self.table_scrips.set_row_bkcolor(row, COLOR_WHITE)
    
    def update_row(self, row):
        self.set_colors(row)
        self.update_pending()
    
    def on_scrips_changed(self, row, col):
        self.data[row][col] = self.table_scrips[row][col]
        self.update_row(row)
        self.modified = True
    
    def header_line(self):
        return "{}".format("".join([
            output_label(index, heading) for index, heading in enumerate(self.headings)
        ]))
    
    def output_lines(self):
        return os.linesep.join(
            [self.header_line()] +
            [scrip.output_line() for scrip in self.data]
        )
    
    def save_scrips(self):
        lines = self.output_lines()
        with open(self.filename, 'w') as f:
            f.writelines(lines)
        self.modified = False
    
    def on_file_save(self):
        self.save_scrips()
    
    def update_state(self, row):
        self.table_scrips[row][:] = self.data[row]
        self.update_row(row)
        self.modified = True
    
    def on_submit(self):
        row = self.table_scrips.current_row()
        if self.data[row].submit():
            self.update_state(row)
        else:
            self.message_box.error("Submit Error", "No refills left.")
    
    def on_refill(self):
        row = self.table_scrips.current_row()
        if self.data[row].refill():
            self.update_state(row)
        else:
            self.message_box.error("Refill Error", "Refill must be submitted first.")
    
    def update_size(self):
        width, height = self.table_scrips.minwidth(), self.table_scrips.minheight()
        self.table_scrips.set_min_size(width, height)
        self.main_window.set_client_area_size(width, height)
    
    def on_add(self):
        scrip = ScripEditable()
        self.data.append(scrip)
        self.table_scrips.append(scrip)
        self.update_pending()
        self.update_size()
        self.modified = True
    
    def on_remove(self):
        scripname = self.data[self.current_row()].name
        msg = "Do you really want to delete {}?".format(scripname)
        if self.message_box.query_ok_cancel("Delete Prescription", msg):
            row = self.table_scrips.current_row()
            del self.table_scrips[row]
            del self.data[row]
            self.update_pending()
            self.update_size()
            self.modified = True
    
    def on_prefs(self):
        self.prefs.show_dialog()
    
    def on_about(self):
        self.about()
    
    def on_about_toolkit(self):
        self.about_toolkit()
    
    def on_exit(self):
        self.exit_app()
    
    def accept_close(self):
        if self.pending:
            if not self.message_box.query_ok_cancel(
                "Pending Submissions",
                "Pending scrips not submitted. Exit anyway?"
            ):
                return False
        if self.modified:
            return self.message_box.query_ok_cancel(
                "Unsaved Changes",
                "Scrips changes not saved. Exit without saving?"
            )
        return True


if __name__ == "__main__":
    ScripListEditor().run()



#!/usr/bin/env python3
"""
Module APP -- UI Application Classes
Sub-Package UI.BASE of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

import sys
import os.path
import contextlib

from plib.stdlib.builtins import inverted
from plib.stdlib.classtools import Singleton
from plib.stdlib.decotools import cached_property

from plib.ui import __version__
from plib.ui.defs import *
from plib.ui.common import *
from plib.ui.imp import get_toolkit_object
from plib.ui.widgets import setup_signal_handlers, widget_from_spec


class PSignalBase(object):
    """Base class for objects that can send signals to notify other objects.
    """
    
    signals = None
    
    def wrap_target(self, signal, target):
        return target
    
    def connect_target(self, signal, target):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def setup_notify(self, signal, target):
        target = self.wrap_target(signal, target)
        self.connect_target(signal, target)
    
    def do_notify(self, signal, *args):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError


class PActionMixin(object):
    """Mixin class for objects that behave like actions.
    """
    
    def get_icon_data(self, key):
        return action_icondata(key)
    
    def load_icon_from_data(self, data):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def load_icon_from_file(self, filename):
        # Most toolkit implementations will probably override this
        # since they provide direct loading of images from files,
        # but this default implementation is here if needed
        with open(filename, 'rb') as f:
            return self.load_icon_from_data(f.read())
    
    def get_icon(self, key):
        return self.load_icon_from_data(self.get_icon_data(key))
    
    def get_menu_str(self, key):
        return action_caption(key)
    
    def get_toolbar_str(self, key):
        return self.get_menu_str(key).replace('&', '').strip('.')
    
    def get_statusbar_str(self, key):
        return action_description(key)
    
    def get_accel_str(self, key):
        return None


class PWidgetBase(PSignalBase):
    """Base class for widgets.
    """
    
    fn_enable_get = None
    fn_enable_set = None
    
    def __init__(self, manager, parent, geometry=None, font=None):
        self.manager = manager
        manager.all_widgets.append(self)
        self.parent = parent
        if geometry is not None:
            self.set_geometry(*geometry)
        if font is not None:
            if isinstance(font, str):
                font = (font,)  # font name only is allowed
            self.set_font(*font)
    
    def setup_signals(self):
        pass  # subclasses can override if needed
    
    def update_widget(self):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def preferred_width(self):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def preferred_height(self):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def get_width(self):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def get_height(self):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def get_size(self):
        return self.get_width(), self.get_height()
    
    def set_size(self, width, height):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def set_width(self, width):
        self.set_size(width, self.get_height())
    
    def set_height(self, height):
        self.set_size(self.get_width(), height)
    
    def set_client_area_size(self, width, height):
        # Factored out so toolkits can override if needed
        self.set_size(width, height)
    
    def get_left(self):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def get_top(self):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def set_position(self, left, top):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def set_min_size(self, width, height):
        """Derived classes must implement.
        """
        raise NotImplementedError
    
    def set_min_width(self, width):
        self.set_min_size(width, self.get_height())
    
    def set_min_height(self, height):
        self.set_min_size(self.get_width(), height)
    
    def set_geometry(self, left, top, width, height):
        """Set widget geometry.
        
        Also sets minimum size; this method is mainly intended to
        be called from the constructor based on parameters given
        in a widget spec.
        """
        if (width is not None) or (height is not None):
            self.set_size(
                width if width is not None else self.get_width(),
                height if height is not None else self.get_height()
            )
            if (width is not None) and (height is not None):
                self.set_min_size(width, height)
            else:
                if width is not None:
                    self.set_min_width(width)
                if height is not None:
                    self.set_min_height(height)
        if (left is not None) or (top is not None):
            self.set_position(
                left if left is not None else self.get_left(),
                top if top is not None else self.get_top()
            )
    
    def set_colors(self, fg=None, bg=None):
        if fg is not None:
            self.set_foreground_color(fg)
        if bg is not None:
            self.set_background_color(bg)
    
    def set_foreground_color(self, color):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def set_background_color(self, color):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def get_font_name(self):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def get_font_size(self):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def get_font_bold(self):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def get_font_italic(self):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def set_font_object(self, font_name, font_size, bold, italic):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def set_font(self, font_name=None,
                 font_size=None, bold=None, italic=None):
        
        self.set_font_object(*font_args(self,
                                        font_name,
                                        font_size,
                                        bold,
                                        italic))
    
    def get_enabled(self):
        return getattr(self, self.fn_enable_get)()
    
    def set_enabled(self, value):
        getattr(self, self.fn_enable_set)(value)
    
    enabled = property(get_enabled, set_enabled)
    
    def set_focus(self):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError


class PAppDialogBase(Singleton):
    
    def _init(self, parent):
        self.parent = parent


class PMessageBoxMeta(type):
    """Metaclass to automatically set up message box classes.
    """
    
    def __init__(cls, name, bases, dict):
        # Add None to question map and set up answer map
        type.__init__(cls, name, bases, dict)
        question_map = getattr(cls, 'question_map')
        question_map.update({ANSWER_NONE: None})
        setattr(cls, 'question_map', question_map)
        answer_map = inverted(question_map)
        setattr(cls, 'answer_map', answer_map)


class PMessageBoxBase(PAppDialogBase, metaclass=PMessageBoxMeta):
    
    question_map = {}
    
    def message_box(self, type, caption, text, default,
                    button1, button2=None, button3=None):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def translate(self, type, caption, text,
                  button1, button2=ANSWER_NONE, button3=ANSWER_NONE):
        
        default = (
            button3 if button3 != ANSWER_NONE else
            button2 if button2 != ANSWER_NONE else
            button1
        )
        return self.answer_map[self.message_box(
            type, caption, text, self.question_map.get(default),
            self.question_map.get(button1),
            self.question_map.get(button2),
            self.question_map.get(button3))]
    
    def info(self, caption, text):
        """Information message box.
        """
        return self.translate(MBOX_INFO, caption, text,
                              ANSWER_OK)
    
    def warn(self, caption, text):
        """Warning message box.
        """
        return self.translate(MBOX_WARN, caption, text,
                              ANSWER_OK)
    
    def error(self, caption, text):
        """Error message box.
        """
        return self.translate(MBOX_ERROR, caption, text,
                              ANSWER_OK)
    
    def query2(self, caption, text, use_yes_no=False):
        """OK/Cancel or Yes/No message box.
        """
        return self.translate(MBOX_QUERY, caption, text,
                              (ANSWER_YES if use_yes_no else ANSWER_OK),
                              (ANSWER_NO if use_yes_no else ANSWER_CANCEL))
    
    def query_ok_cancel(self, caption, text):
        """Returns True for Ok and False for Cancel.
        """
        return (self.query2(caption, text) == ANSWER_OK)
    
    def query_yes_no(self, caption, text):
        """Returns True for Yes and False for No.
        """
        return (self.query2(caption, text, use_yes_no=True) == ANSWER_YES)
    
    def query3(self, caption, text):
        """Yes/No/Cancel message box.
        """
        return self.translate(MBOX_QUERY, caption, text,
                              ANSWER_YES, ANSWER_NO, ANSWER_CANCEL)
    
    def query_yes_no_cancel(self, caption, text):
        """Returns True for Yes, False for No, None for Cancel
        """
        answer = self.query3(caption, text)
        return (
            True if answer == ANSWER_YES else
            False if answer == ANSWER_NO else
            None
        )
    
    def enter_value(self, value_type, caption, prompt):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def enter_string(self, caption, prompt):
        return self.enter_value(str, caption, prompt)
    
    def enter_int(self, caption, prompt):
        return self.enter_value(int, caption, prompt)


class PFileDialogBase(PAppDialogBase):
    
    def choose_directory(self, curdir):
        """Display select folder dialog and return chosen folder.
        """
        raise NotImplementedError
    
    def _translate_filter_item(self, caption, ext_str):
        """Translate filter item into toolkit format.
        """
        raise NotImplementedError
    
    ext_sep = None
    
    def _ext_str_from_list(self, ext_list):
        return self.ext_sep.join("*.{}".format(ext) for ext in ext_list)
    
    def _translate_filter_listitem(self, caption, ext_list):
        return self._translate_filter_item(caption, self._ext_str_from_list(ext_list))
    
    filter_sep = None
    
    def _translate_filter(self, filter):
        return self.filter_sep.join(
            self._translate_filter_listitem(caption, ext_list)
            for caption, ext_list in filter
        )
    
    def _open_filename(self, path, filter, selected_filter):
        """Display select file dialog and return chosen filename for opening.
        """
        raise NotImplementedError
    
    def _save_filename(self, path, filter, selected_filter):
        """Display select file dialog and return chosen filename for saving.
        """
        raise NotImplementedError
    
    def open_filename(self, path, filter, selected_filter):
        return self._open_filename(path, self._translate_filter(filter), self._translate_filter_listitem(*selected_filter))
    
    def save_filename(self, path, filter, selected_filter):
        return self._save_filename(path, self._translate_filter(filter), self._translate_filter_listitem(*selected_filter))


class PAboutDialogBase(PAppDialogBase):
    """Base class for about dialogs.
    """
    
    plibstr = "Using PLIB.UI version {}".format(__version__)
    
    display_func = None
    
    def display(self):
        data = self.parent.app.about_data
        fmt = self.parent.app.about_format
        caption = "About {}".format(data.get('name', "Application"))
        aboutstr = fmt.format(**data)
        body = "{}\n\n{}".format(aboutstr, self.plibstr)
        self.display_func(self.parent, caption, body)


class PProgressDialogBase(PAppDialogBase):
    """Base class for progress dialogs.
    """
    
    def create_dialog(self, title, msg, max_num, stop_label, main_window):
        """Create progress dialog as modal child of main_window
        """
        raise NotImplementedError
    
    def show_dialog(self):
        """Show dialog immediately, before yielding in context manager.
        """
        raise NotImplementedError
    
    def was_canceled(self):
        """Return whether progress dialog was canceled
        """
        raise NotImplementedError
    
    def update_progress(self, num):
        """Update progress dialog status
        """
        raise NotImplementedError
    
    def complete(self):
        """Progress dialog is complete.
        """
        raise NotImplementedError
    
    def close_dialog(self):
        """Close progress dialog.
        """
        raise NotImplementedError
    
    @contextlib.contextmanager
    def updater(self, title, msg, max_num, stop_label=None):
        """Show progress dialog and allow updating.
        """
        
        self.dialog = self.create_dialog(title, msg, max_num, stop_label, self.parent)
        try:
            self.show_dialog()
            yield self
        finally:
            self.close_dialog()
            del self.dialog


class PTopWindowBase(PWidgetBase):
    """Base class for 'top window' widgets.
    
    A top window is a 'plain' main application window; it has no
    frills like menus, toolbars, status bars, etc. built in.
    """
    
    signals = (
        SIGNAL_SHOWN,
        SIGNAL_QUERYCLOSE,
        SIGNAL_CLOSING,
        SIGNAL_HIDDEN,
    )
    
    def __init__(self, manager):
        PWidgetBase.__init__(self, manager, None)  # no parent
        self.shown = False
        self.client_widget = None
        
        setup_signal_handlers(self.app, self, 'app')
    
    @property
    def app(self):
        return self.manager  # these are always the same for a top window
    
    def before_client(self):
        """For subclasses, to allow initialization before client widget is created
        """
        pass
    
    def set_client_widget(self, client_widget):
        self.client_widget = client_widget
    
    def set_caption(self, caption):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def get_desktop_rect(self, primary=True):
        """Return 6-tuple (index, num_desktops, left, top, width, height) for desktop.
        """
        raise NotImplementedError
    
    def get_client_size(self):
        """Return tuple of (width, height) needed to wrap client.
        """
        c = self.client_widget
        return (c.preferred_width(), c.preferred_height())
    
    def size_to_client(self):
        self.set_client_area_size(*self.get_client_size())
    
    def size_maximize(self):
        """Size the window to be maximized.
        """
        raise NotImplementedError
    
    def size_to_screen(self, width_offset=0, height_offset=0):
        l, t, w, h = self.get_desktop_rect()
        self.set_size(w - width_offset, h - height_offset)
        self.set_position(l, t)
    
    def get_frame_size(self):
        """Return tuple of (width, height) for the window's frame size.
        """
        raise NotImplementedError
    
    def center(self):
        l, t, w, h = self.get_desktop_rect()
        x, y = self.get_frame_size()
        self.set_position(l + (w - x) / 2, t + (h - y) / 2)
    
    def init_placement(self, size, pos):
        if size == SIZE_CLIENTWRAP:
            self.size_to_client()
        elif size == SIZE_MAXIMIZED:
            self.size_maximize()
        elif size == SIZE_DESKTOP:
            self.size_to_screen()
        elif size == SIZE_OFFSET:
            self.size_to_screen(self.app.width_offset, self.app.height_offset)
        if (pos == MOVE_CENTER) and (size != SIZE_MAXIMIZED):
            self.center()
    
    def show_init(self):
        """Should always call from derived classes to ensure proper setup.
        """
        if not self.shown:
            # Do placement just before showing for first time
            self.init_placement(self.app.main_size, self.app.main_placement)
            self.shown = True
    
    def can_close(self):
        # To be called after SIGNAL_QUERYCLOSE is handled, so handlers
        # for that signal can adjust any state to be checked here
        return self.app.accept_close()
    
    def do_exit(self):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def exit(self):
        """Ensure client references are removed before exiting.
        """
        self.client_widget = None
        self.do_exit()
    
    def set_iconfile(self, iconfile):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def get_current_geometry(self):
        """Return the window's current geometry (including frame).
        """
        raise NotImplementedError


class PManagerBase:
    """Base class for non-widget object that manages widgets.
    
    Objects of this type can be the ``manager`` in a widget's constructor. The main
    function of widget managers is to implement event handlers for the widgets
    they manage.
    """
    
    def __init__(self):
        self.all_widgets = []
        self.sub_managers = []
    
    # Note: every non-abstract subclass of this class must either initialize the manager attribute
    # in its constructor, or override the app and controller properties (as PApplication does,
    # to just return self)
    
    @property
    def app(self):
        app = self.manager
        while not isinstance(app, PApplicationBase):
            app = app.manager
        return app
    
    @property
    def controller(self):
        controller = self.manager
        while not isinstance(controller, PControllerBase):
            controller = controller.manager
        return controller
    
    handler_prefix = None
    
    def setup_signals(self):
        for widget in self.all_widgets:
            widget.setup_signals()
        for manager in self.sub_managers:
            manager.setup_signals()
        if self.handler_prefix:
            setup_signal_handlers(self, self, self.handler_prefix)
    
    def do_after_create(self):
        for manager in self.sub_managers:
            manager.do_after_create()
        self.after_create()
    
    def do_populate_data(self):
        for manager in self.sub_managers:
            manager.check_populate_data()
        self.check_populate_data()
    
    showing = True
    
    def check_populate_data(self):
        if self.showing:
            self.populate_data()
    
    def after_create(self):
        """Placeholder for derived classes if needed.
        
        This method can assume that all widgets are created and available and that
        all auto-connected signal handlers are connected.
        """
        pass
    
    def populate_data(self):
        """Placeholder for derived classes if needed.
        
        This method can assume that all widgets are created and available, that
        all auto-connected signal handlers are connected, and that all after_create
        processing has been completed.
        """
        pass


class PControllerBase(PManagerBase):
    """Base class for widget controllers.
    
    Widget controllers are widget managers, but they also control the actual widget
    creation process. Currently the only controllers are the application object and
    dialog runners.
    """
    
    def before_create(self):
        """Placeholder for derived classes if needed.
        
        This method is called before any widgets are created or specs are evaluated.
        Note that we cannot call any widgets or sub-managers because they have not
        been created or initialized yet. That is why this method is only on the app,
        not any other widget managers.
        """
        pass
    
    def create_widgets(self):
        """This method should create whatever widgets are being managed.
        
        Derived classes must override and implement.
        """
        raise NotImplementedError
    
    def do_create(self):
        """This method should be called at the proper time to do widget creation.
        """
        
        # Allow custom processing before any widgets are created
        self.before_create()
        
        # Create the widgets
        self.create_widgets()
        
        # This is factored out to allow custom signal setup by toolkits if needed;
        # we do it before do_after_create to ensure that there are not multiple
        # references to widgets or managers
        self.setup_signals()
        
        # Allow custom processing after all widgets are created
        self.do_after_create()
        
        # Allow loading of data after everything else is done
        self.do_populate_data()


class PApplicationBase(PControllerBase, PSignalBase):
    """Base class for GUI application.
    """
    
    signals = ()
    
    handler_prefix = 'app'
    
    about_dialog_class = None
    message_box_class = None
    file_dialog_class = None
    progress_dialog_class = None
    
    about_toolkit_func = None
    
    about_data = {}
    about_format = None
    
    main_title = "Application"
    main_iconfile = None
    main_size = SIZE_NONE
    main_placement = MOVE_NONE
    
    width_offset = height_offset = 160
    
    main_widget = None
    
    menu_actions = None
    toolbar_actions = None
    
    include_statusbar = True
    
    large_icons = False
    show_labels = False
    
    status_labels = None
    
    def __init__(self, arglist=[]):
        PControllerBase.__init__(self)
        self.arglist = arglist
    
    @property
    def app(self):
        return self  # for compatibility with other widget managers/controllers
    
    @property
    def controller(self):
        return self  # for compatibility with other widget managers/controllers
    
    def create_widgets(self):
        self.main_window = self.create_main_widget()
        self.main_window.before_client()
        self.client_widget = self.create_client_widget()
        if self.main_iconfile:
            self.main_window.set_iconfile(self.main_iconfile)
    
    @property
    def use_mainwindow(self):
        return bool(self.menu_actions or self.toolbar_actions)
    
    @property
    def include_menu(self):
        return bool(self.menu_actions)
    
    @property
    def include_toolbar(self):
        return bool(self.toolbar_actions)
    
    def create_main_widget(self):
        """Create the main widget and return it.
        """
        
        args = (
            ('mainwin', 'PMainWindow') if self.use_mainwindow else
            ('app', 'PTopWindow')
        )
        main_klass = get_toolkit_object(*args)
        main_window = main_klass(self)
        main_window.set_caption(self.main_title)
        return main_window
    
    def create_client_widget(self):
        spec = self.main_widget
        if spec:
            main_window = self.main_window
            client_widget = widget_from_spec(self, main_window, spec)
            main_window.set_client_widget(client_widget)
            return client_widget
        return None
    
    @cached_property
    def message_box(self):
        return self.message_box_class(self.main_window)
    
    @cached_property
    def file_dialog(self):
        return self.file_dialog_class(self.main_window)
    
    @cached_property
    def about_dialog(self):
        return self.about_dialog_class(self.main_window)
    
    @cached_property
    def progress_dialog(self):
        return self.progress_dialog_class(self.main_window)
    
    def about(self):
        self.about_dialog.display()
    
    def about_toolkit(self):
        self.about_toolkit_func()
    
    def event_loop(self):
        """Placeholder for derived classes for main event loop.
        """
        raise NotImplementedError
    
    def run(self):
        """Show the main widget and run the main event loop.
        """
        
        self.main_window.show_init()
        self.event_loop()
        self.main_window = None
    
    def process_events(self):
        """Placeholder for derived classes to pump events outside main loop.
        """
        raise NotImplementedError
    
    def exit_app(self):
        self.main_window.exit()
    
    def accept_close(self):
        """Return False if app should not close based on current state.
        """
        return True
    
    def before_quit(self):
        """Placeholder for derived classes if needed.
        
        Note that this method cannot assume that *any* objects other than
        the application itself are still available; things that need to be
        done with widgets still available should be done in the accept_close
        method above.
        """
        pass



#!/usr/bin/env python3
"""
Module SCROLLER -- UI Scroller Widgets
Sub-Package UI.BASE of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

from plib.ui.widgets import widget_from_spec

from .app import PWidgetBase


class PScrollerBase(PWidgetBase):
    
    def __init__(self, manager, parent, child=None):
        PWidgetBase.__init__(self, manager, parent)
        if child:
            self.set_child(child)
    
    def set_child(self, child):
        """Derived classes must implement.
        """
        raise NotImplementedError


class PScrollingWrapperBase(PScrollerBase):
    
    def __init__(self, manager, parent, child_spec):
        PScrollerBase.__init__(self, manager, parent)
        child = widget_from_spec(manager, self.spec_parent, child_spec)
        layout = self.get_layout(child)
        if layout is not None:
            self.setup_layout(layout)
        self.set_child(child)
    
    @property
    def spec_parent(self):
        return self.parent
    
    def get_layout(self, child):
        return None  # Derived classes can override
    
    def setup_layout(self, layout):
        pass  # Derived classes can override



#!/usr/bin/env python3
"""
Module FOCUS -- UI Focus Widget Mixin
Sub-Package UI.TOOLKITS.WX of Package PLIB3 -- Python GUI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

from plib.ui.defs import *
from plib.ui.base.focus import PFocusBase


class PFocusMixin(PFocusBase):
    pass  # no special code required here since the focus events are built in to all wx widgets



#!/usr/bin/env python3
"""
Module CONTAINER -- UI Container Widget
Sub-Package UI.BASE of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""


from .app import PWidgetBase


class PContainerBase(PWidgetBase):
    
    child = None
    
    def add_child(self, widget):
        if self.child:
            raise RuntimeError("Can't add more than one child to a container.")
        self.child = widget
        self.add_widget(widget)
        self.do_layout()
    
    def add_widget(self, widget):
        # Derived classes may override to do any necessary
        # adjustment when a widget is added.
        pass
    
    def do_layout(self):
        # Derived classes should override to do any
        # necessary creation of physical layout objects
        # after all child panels are created.
        pass



#!/usr/bin/env python3
"""
Module LISTBOX-- UI List Box Widget
Sub-Package UI.BASE of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

import functools

from plib.stdlib.coll import SortMixin

from plib.ui.defs import *
from plib.ui.coll import BaseStringListWidget

from .app import PWidgetBase


class PListBoxBase(PWidgetBase, BaseStringListWidget):
    """List box that looks like a list of strings.
    """
    
    signals = (
        SIGNAL_LISTBOXSELECTED,
    )
    
    default_itemheight = 30
    
    def __init__(self, manager, parent, items=None, value=None, starting_index=None,
                 geometry=None, font=None):
        
        PWidgetBase.__init__(self, manager, parent,
                             geometry=geometry, font=font)
        BaseStringListWidget.__init__(self, items, value, starting_index)
    
    def item_height(self, index):
        return self.default_itemheight
    
    def minheight(self):
        return sum(self.item_height(index) for index in range(len(self)))


class PSortedListBoxBase(SortMixin, PListBoxBase):
    """List box that automatically sorts its items.
    """
    
    def __init__(self, manager, parent, items=None, value=None, starting_index=None,
                 geometry=None, font=None, key=None):
        
        PListBoxBase.__init__(self, manager, parent, geometry=geometry, font=font)  # don't pass items here
        self._init_seq(items, key)  # sort items and add them here
        self.complete_init(items, value, starting_index)  # do this since the inherited constructor didn't see items or value



#!/usr/bin/env python3
"""
Module FOCUS -- UI Focus Widget Mixin
Sub-Package UI.TOOLKITS.PYSIDE2 of Package PLIB3 -- Python UI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

from plib.ui.defs import *
from plib.ui.base.focus import PFocusBase

from .app import PQtWidgetMeta


class PFocusMixin(PFocusBase, metaclass=PQtWidgetMeta):
    
    event_signals = (
        SIGNAL_FOCUS_IN,
        SIGNAL_FOCUS_OUT
    )
    
    def focusInEvent(self, event):
        super(PFocusMixin, self).focusInEvent(event)
        self.do_notify(SIGNAL_FOCUS_IN)
    
    def focusOutEvent(self, event):
        super(PFocusMixin, self).focusOutEvent(event)
        self.do_notify(SIGNAL_FOCUS_OUT)



#!/usr/bin/env python3
"""
Module WX.COMBO -- Python wxWidgets Combo Box Widgets
Sub-Package UI.TOOLKITS.WX of Package PLIB3 -- Python GUI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the wxWidgets UI objects for combo boxes.
"""

import wx

from plib.ui.defs import *
from plib.ui.base.combo import PComboBoxBase, PNumComboBoxBase, PSortedComboBoxBase

from .app import PWxSequenceWidget


class PWxComboBoxBase(PWxSequenceWidget, wx.ComboBox):
    
    _align = False  # used by panel to determine placement
    _expand_horiz = True
    
    def __init__(self, parent):
        wx.ComboBox.__init__(self, parent,
                             style=(wx.CB_DROPDOWN | wx.CB_READONLY))
    
    def current_text(self):
        return self.GetStringSelection()
    
    def set_current_text(self, text):
        self.SetStringSelection(text)
    
    def current_index(self):
        return self.GetSelection()
    
    def set_current_index(self, index):
        self.SetSelection(index)
    
    def _indexlen(self):
        return self.GetCount()
    
    def _get_data(self, index):
        return self.GetString(index)
    
    def _set_data(self, index, value):
        self.SetString(index, value)
    
    def _add_data(self, index, value):
        if index == self.__len__():
            self.Append(value)
        else:
            self.Insert(value, index)
    
    def _del_data(self, index):
        self.Delete(index)


class PComboBox(PWxComboBoxBase, PComboBoxBase):
    
    def __init__(self, manager, parent, items=None, value=None,
                 geometry=None, font=None):
        
        PWxComboBoxBase.__init__(self, parent)
        PComboBoxBase.__init__(self, manager, parent, items, value=value,
                               geometry=geometry, font=font)


class PNumComboBox(PWxComboBoxBase, PNumComboBoxBase):
    
    def __init__(self, manager, parent, items=None, value=None,
                 geometry=None, font=None):
        
        PWxComboBoxBase.__init__(self, parent)
        PNumComboBoxBase.__init__(self, manager, parent, items, value=value,
                                  geometry=geometry, font=font)


class PSortedComboBox(PWxComboBoxBase, PSortedComboBoxBase):
    
    def __init__(self, manager, parent, items=None, value=None,
                 geometry=None, font=None, key=None):
        
        PWxComboBoxBase.__init__(self, parent)
        PSortedComboBoxBase.__init__(self, manager, parent, items, value=value,
                                     geometry=geometry, font=font, key=key)
